#!/usr/bin/perl -w
#
#  planets.nu interface
#
#  This script accesses the Nu server by HTTP and converts between Nu
#  data and VGA Planets data. It operates in the current directory. It
#  will maintain a state file which stores some settings as well as
#  required cookies.
#
#  A Nu RST file is a JSON data structure and contains the
#  specification files, RST, and some history data.
#
#  Usage:
#    perl c2nu.pl [--api=H] [--backups=[01]] [--root=DIR] CMD [ARGS...]
#
#  Options:
#    --api=H        set API host name (setting will be stored in state file)
#    --backups=0/1  disable/enable backup of received Nu RST files. Note
#                   that those are rather big so it makes sense to compress
#                   the backups.
#    --dropmines=0/1  disable/enable removal of mines from PCC chart db
#    --root=DIR     set root directory containing VGA Planets spec files.
#                   Those are used to "fill in the blanks" not contained
#                   in a Nu RST.
#
#  Commands:
#    help           Help screen (no network access)
#    status         Show state file content (no network access)
#    login USER PW  Log in with user Id and password
#    list           List games (must be logged in)
#    vcr [GAME]     Download Nu RST and create VGAP VCRx.DAT for PlayVCR.
#
#  All download commands can be split in two halves, i.e. "vcr1 [GAME]"
#  to perform the download, and "vcr2" to convert the download without
#  accessing the network.
#
#  Instructions:
#  - make a directory and go there using the command prompt
#  - log in using 'c2nu login USER PASS'
#  - list games using 'c2nu list'
#  - download stuff using 'c2nu --root=DIR vcr', where DIR is the
#    directory containing your VGA Planets installation, or PCC2's
#    'specs' direcory. Alternatively, copy a 'hullfunc.dat' file
#    into the current directory before downloading the game.
#
#  Since the server usually sends gzipped data, this script needs the
#  'gzip' program in the path to decompress it.
#
#  (c) 2011-2012,2016 Stefan Reuther
#
use strict;
use Socket;
use IO::Handle;
use bytes;              # without this, perl 5.6.1 doesn't correctly read Unicode stuff

my $VERSION = "0.3";
my $opt_rootDir = "/usr/share/planets";

# Initialisation
stateSet('api', 'api.planets.nu');
stateLoad();

# Parse arguments
while (@ARGV) {
    if ($ARGV[0] =~ /^--?api=(.*)/) {
        stateSet('api', $1);
        shift @ARGV;
    } elsif ($ARGV[0] =~ /^--?backup=(\d+)/) {
        stateSet('backups', $1);
        shift @ARGV;
    } elsif ($ARGV[0] =~ /^--?drop[-_]?mines=(\d+)/) {
        stateSet('dropmines', $1);
        shift @ARGV;
    } elsif ($ARGV[0] =~ /^--?root=(.*)/) {
        $opt_rootDir = $1;
        shift @ARGV;
    } else {
        last;
    }
}

# Command switch
if (!@ARGV) {
    die "Missing command name. '$0 --help' for help.\n";
}
my $cmd = shift @ARGV;
$cmd =~ s/^--?//;
if ($cmd eq 'help') {
    doHelp();
} elsif ($cmd eq 'status') {
    doStatus();
} elsif ($cmd eq 'login') {
    doLogin();
} elsif ($cmd eq 'list') {
    doList();
} elsif ($cmd =~ /^vcr([12]?)$/) {
    doDownloadResult() unless $1 eq '2';
    doWriteVcr()       unless $1 eq '1';
} else {
    die "Invalid command '$cmd'. '$0 --help' for help\n";
}
stateSave();
exit 0;

######################################################################
#
#  Help
#
######################################################################
sub doHelp {
    print "$0 - planets.nu interface - version $VERSION, (c) 2011-2012,2016 Stefan Reuther\n\n";
    print "$0 [options] command [command args]\n\n";
    print "Options:\n";
    print "  --api=HOST        instead of 'api.planets.nu'\n";
    print "  --backups=0/1     disable/enable backup of Nu RST\n";
    print "  --dropmines=0/1   disable/enable removal of minefields from chart DB\n";
    print "  --root=DIR        set root directory containing VGA Planets spec files\n\n";
    print "Commands:\n";
    print "  help              this help screen\n";
    print "  status            show status\n";
    print "  login USER PASS   log in\n";
    print "  list              list games\n";
    print "  vcr [GAME]        download Nu RST and create VGAP VCRx.DAT\n";
    print "\n";
    print "Download commands can be split into the download part ('vcr1') and the\n";
    print "convert part ('vcr2').\n";
}

######################################################################
#
#  Log in
#
######################################################################
sub doLogin {
    if (@ARGV != 2) {
        die "login: need two arguments, user name and password\n";
    }

    my $user = $ARGV[0];
    my $pass = $ARGV[1];

    my $reply = httpCall("POST /account/login?version=2 HTTP/1.0\n",
                         httpBuildQuery(username => $user,
                                        password => $pass));

    my $parsedReply = jsonParse($reply->{BODY});
    if (exists($parsedReply->{success}) && ($parsedReply->{success} =~ /true/i || $parsedReply->{success})) {
        print "++ Login succeeded ++\n";
        stateSet('user', $user);
        stateSet('apikey', $parsedReply->{apikey});
        foreach (sort keys %$reply) {
            if (/^COOKIE-(.*)/) {
                stateSet("cookie_$1", $reply->{$_});
            }
        }
    } else {
        print "++ Login failed ++\n";
        print "Server answer:\n";
        foreach (sort keys %$parsedReply) {
            printf "%-20s %s\n", $_, $parsedReply->{$_};
        }
    }
}

######################################################################
#
#  List
#
######################################################################
sub doList {
    my $reply = httpCall("POST /account/mygames?version=2 HTTP/1.0\n", httpBuildQuery(apikey => stateGet('apikey')));
    my $parsedReply = jsonParse($reply->{BODY});
    my $needHeader = 1;
    if (exists($parsedReply->{games})) {
        foreach (@{$parsedReply->{games}}) {
            my $gameName = $_->{game}{name} || '?';
            my $gameNr   = $_->{game}{id}   || 0;
            my $type     = $_->{game}{gametype} == 2 ? 'Normal' :
                $_->{game}{gametype} == 1 ? 'Training' : '?';
            my $race = $_->{player}{raceid} || '?';

            # Print
            print "Game      Name                                      Race  Category\n" if $needHeader;
            print "--------  ----------------------------------------  ----  --------------------\n" if $needHeader;
            printf "%8d  %-40s  %4d  %s\n", $gameNr, $gameName, $race, $type;
            $needHeader = 0;
        }
    } else {
        print "++ Unable to obtain game list ++\n";
    }
}


######################################################################
#
#  VCR file
#
######################################################################

sub doWriteVcr {
    # Read state
    my $body = readFile("c2rst.txt");
    print "Parsing result...\n";
    doVcr(jsonParse($body));
}

sub doVcr {
    # Fetch parameter
    my $parsedReply = shift;
    stateCheckReply($parsedReply);

    # Make spec files
    makeAllSpecFiles($parsedReply);

    # Make result
    my $player = $parsedReply->{rst}{player}{id};
    my $vcrs = rstPackVcrs($parsedReply, $player);
    my $fn = "vcr$player.dat";
    open VCR, "> $fn" or die "$fn: $!\n";
    print "Making $fn...\n";
    binmode VCR;
    print VCR $vcrs;
    close VCR;
    if (length($vcrs) < 100) {
        print "++ You do not have any VCRs this turn. ++\n";
    }
}


######################################################################
#
#  Result file
#
######################################################################

sub doDownloadResult {
    my $gameId;
    if (@ARGV == 0) {
        $gameId = stateGet('gameid');
        if (!$gameId) {
            die "rst1: need one parameter: game name\n";
        }
    } elsif (@ARGV == 1) {
        $gameId = shift @ARGV;
    } else {
        die "rst1: need one parameter: game name\n";
    }
    stateSet('gameid', $gameId);

    my $reply = httpCall("POST /game/loadturn HTTP/1.0\n",
                         httpBuildQuery(gameid => $gameId,
                                        apikey => stateGet('apikey'),
                                        forsave => "true",
                                        activity => "true"));

    print "Saving output...\n";
    open OUT, "> c2rst.txt" or die "c2rst.txt: $!\n";
    print OUT $reply->{BODY};
    close OUT;

    print "Parsing result...\n";
    my $parsedReply = jsonParse($reply->{BODY});
    if (!exists $parsedReply->{rst}) {
        print STDERR "WARNING: request probably did not succeed.\n";
        if (exists $parsedReply->{error}) {
            print STDERR "WARNING: error message is:\n\t", $parsedReply->{error}, "\n";
        }
    } else {
        if (stateGet('backups')) {
            print "Making backup...\n";
            my $turn = $parsedReply->{rst}{settings}{turn};
            my $player = $parsedReply->{rst}{player}{raceid};
            open OUT, sprintf("> c2rst_backup_%d_%03d.txt", $player, $turn);
            print OUT $reply->{BODY};
            close OUT;
        }
    }
}

# Make all specification files
sub makeAllSpecFiles {
    my $parsedReply = shift;

    # Simple spec files
    makeSpecFile($parsedReply->{rst}{beams}, "beamspec.dat", 10, "A20v8", 36,
                 qw(name cost tritanium duranium molybdenum mass techlevel crewkill damage));
    makeSpecFile($parsedReply->{rst}{torpedos}, "torpspec.dat", 10, "A20v9", 38,
                 qw(name torpedocost launchercost tritanium duranium molybdenum mass techlevel crewkill damage));
    makeSpecFile($parsedReply->{rst}{engines}, "engspec.dat", 9, "A20v5V9", 66,
                 qw(name cost tritanium duranium molybdenum techlevel warp1 warp2 warp3 warp4 warp5 warp6 warp7 warp8 warp9));
    makeSpecFile($parsedReply->{rst}{hulls}, "hullspec.dat", 105, "A30v15", 60,
                 qw(name zzimage zzunused tritanium duranium molybdenum fueltank
                    crew engines mass techlevel cargo fighterbays launchers beams cost));
    makeSpecFile($parsedReply->{rst}{planets}, "xyplan.dat", 500, "v3", 6,
                 qw(x y ownerid));
    makeSpecFile($parsedReply->{rst}{planets}, "planet.nm", 500, "A20", 20,
                 qw(name));

    # Make more spec files
    makeHullfuncFile($parsedReply->{rst}{hulls});
    makeTruehullFile($parsedReply->{rst}{racehulls}, $parsedReply->{rst}{player}{raceid});
}

# Make specification file from data received within nu RST
sub makeSpecFile {
    my $replyPart = shift;
    my $fileName = shift;
    my $numEntries = shift;
    my $packPattern = shift;
    my $entrySize = shift;
    my @fields = @_;

    print "Making $fileName...\n";

    # Build field-to-slot mapping and entry template
    my %fieldToSlot;
    my @entryTemplate;
    foreach (0 .. $#fields) {
        $fieldToSlot{$fields[$_]} = $_;
        push @entryTemplate, 0;
    }

    # Load existing file or build empty file
    my @file;
    if ($fileName eq 'xyplan.dat') {
        # do NOT create xyplan.dat from original, it is expected to have holes!
        foreach (1 .. $numEntries) {
            push @file, [0,0,0];
        }
    } elsif (open(FILE, "< $fileName") || open(FILE, "< $opt_rootDir/$fileName")) {
        binmode FILE;
        foreach (1 .. $numEntries) {
            my $buf;
            read FILE, $buf, $entrySize;
            push @file, [unpack $packPattern, $buf];
        }
        close FILE;
    } else {
        if ($fileName eq 'hullspec.dat') {
            print "WARNING: 'hullspec.dat' created from scratch; it will not contain image references.\n";
            print "    Copy a pre-existing 'hullspec.dat' into this directory and process the RST again\n";
            print "    to have images.\n";
        }
        foreach (1 .. $numEntries) {
            if (exists $fieldToSlot{name}) { $entryTemplate[$fieldToSlot{name}] = "#$_"; }
            push @file, [@entryTemplate];
        }
    }

    # Populate file
    foreach my $e (@$replyPart) {
        if ($e->{id} > 0 && $e->{id} <= $numEntries) {
            foreach (sort keys %$e) {
                if (exists $fieldToSlot{$_}) {
                    $file[$e->{id} - 1][$fieldToSlot{$_}] = $e->{$_};
                }
            }
        }
    }

    # Generate it
    open FILE, "> $fileName" or die "$fileName: $!\n";
    binmode FILE;
    foreach (@file) {
        print FILE pack($packPattern, @$_);
    }
    close FILE;
}

sub makeHullfuncFile {
    # Nu stores cloakiness in its spec file, but all other hull
    # functions appear as free-form text only. We therefore assume
    # the hull functions to be reasonably default, and generate
    # a hullfunc file which only updates the Cloak ability.
    # FIXME: what about AdvancedCloak?
    my $hulls = shift;
    print "Making hullfunc.txt...\n";
    open FILE, "> hullfunc.txt" or die "hullfunc.txt: $!\n";
    print FILE "# Hull function definitions for 'nu' game\n\n";
    print FILE "\%hullfunc\n\n";
    print FILE "Init = Default\n";
    print FILE "Function = Cloak\n";
    print FILE "Hull = *\n";
    print FILE "RacesAllowed = -\n";
    foreach (@$hulls) {
        if ($_->{cancloak}) {
            print FILE "Hull = ", $_->{id}, "\n";
            print FILE "RacesAllowed = +\n";
        }
    }
    close FILE;
}

sub makeTruehullFile {
    my $pRacehulls = shift;
    my $player = shift;

    print "Making truehull.dat...\n";

    # Load existing file if any
    my @truehull = replicate(20*11, 0);
    if (open(TH, "< truehull.dat") or open(TH, "< $opt_rootDir/truehull.dat")) {
        my $th;
        binmode TH;
        read TH, $th, 20*11*2;
        close TH;
        @truehull = unpack("v*", $th);
    }

    # Merge race hulls
    for (my $i = 0; $i < 20; ++$i) {
        $truehull[($player-1)*20 + $i] = ($i < @$pRacehulls ? $pRacehulls->[$i] : 0);
    }

    # Write
    open(TH, "> truehull.dat") or die "truehull.dat: $!\n";
    binmode TH;
    print TH pack("v*", @truehull);
    close TH;
}

######################################################################
#
#  RST creation
#
######################################################################

sub rstPackVcrs {
    my $parsedReply = shift;
    my $player = shift;
    my @vcrs;
    foreach my $vcr (@{$parsedReply->{rst}{vcrs}}) {
        my $v = pack("v*",
                     $vcr->{seed},
                     0x554E,       # 'NU', signature
                     $vcr->{right}{temperature},
                     $vcr->{battletype},
                     $vcr->{left}{mass},
                     $vcr->{right}{mass});
        foreach (qw(left right)) {
            my $o = $vcr->{$_};
            $v .= pack("A20v11",
                       $o->{name},
                       $o->{damage},
                       $o->{crew},
                       $o->{objectid},
                       $o->{raceid},
                       256*$o->{hullid} + 1,  # image, hull
                       $o->{beamid},
                       $o->{beamcount},
                       $o->{baycount},
                       $o->{torpedoid},
                       $o->{baycount}==0 ? $o->{torpedos} : $o->{fighters},
                       $o->{launchercount});
        }
        $v .= pack("vv", $vcr->{left}{shield}, $vcr->{right}{shield});
        push @vcrs, $v;
    }
    pack("v", scalar(@vcrs)) . join("", @vcrs);
}


######################################################################
#
#  State file
#
######################################################################

my %stateValues;
my %stateChanged;

sub stateLoad {
    if (open(STATE, "< c2nu.ini")) {
        while (<STATE>) {
            s/[\r\n]*$//;
            next if /^ *#/;
            next if /^ *$/;
            if (/^(.*?)=(.*)/) {
                my $key = $1;
                my $val = $2;
                $val =~ s|\\(.)|stateUnquote($1)|eg;
                $stateValues{$key} = $val;
                $stateChanged{$key} = 0;
            } else {
                print "WARNING: state file line $. cannot be parsed\n";
            }
        }
        close STATE;
    }
}

sub stateSave {
    # Needed?
    my $needed = 0;
    foreach (keys %stateValues) {
        if ($stateChanged{$_}) {
            # print "Must update state file because '$_' has changed.\n";
            $needed = 1;
            last;
        }
    }
    return if !$needed;
    print "Updating state file...\n";

    # Copy existing file, updating it
    open(OUT, "> c2nu.new") or die "ERROR: cannot create new state file c2nu.new: $!\n";
    if (open(STATE, "< c2nu.ini")) {
        while (<STATE>) {
            s/[\r\n]*$//;
            if (/^ *#/ || /^ *$/) {
                print OUT "$_\n";
            } elsif (/^(.*?)=(.*)/ && $stateChanged{$1}) {
                my $key = $1;
                print OUT "$key=", stateQuote($stateValues{$key}), "\n";
                $stateChanged{$key} = 0;
            } else {
                print OUT "$_\n";
            }
        }
        close STATE;
    }

    # Print missing keys
    foreach (sort keys %stateValues) {
        if ($stateChanged{$_}) {
            print OUT "$_=", stateQuote($stateValues{$_}), "\n";
            $stateChanged{$_} = 0;
        }
    }
    close OUT;

    # Rename files
    unlink "c2nu.bak";
    rename "c2nu.ini", "c2nu.bak";
    rename "c2nu.new", "c2nu.ini" or print "WARNING: cannot rename new state file: $!\n";
}

sub stateSet {
    my $key = shift;
    my $val = shift;
    if (!exists($stateValues{$key}) || $stateValues{$key} ne $val) {
        $stateValues{$key} = $val;
        $stateChanged{$key} = 1;
    }
}

sub stateGet {
    my $key = shift;
    if (exists($stateValues{$key})) {
        $stateValues{$key}
    } else {
        "";
    }
}

sub stateQuote {
    my $x = shift;
    $x =~ s/\\/\\\\/g;
    $x =~ s/\n/\\n/g;
    $x =~ s/\r/\\r/g;
    $x =~ s/\t/\\t/g;
    $x =~ s/\t/\\t/g;
    $x =~ s/"/\\"/g;
    $x =~ s/'/\\'/g;
    $x;
}

sub stateUnquote {
    my $x = shift;
    if ($x eq 'n') {
        return "\n";
    } elsif ($x eq 't') {
        return "\t";
    } elsif ($x eq 'r') {
        return "\r";
    } else {
        return $x;
    }
}

sub stateCookies {
    my @cookie;
    foreach (sort keys %stateValues) {
        if (/^cookie_(.*)/) {
            push @cookie, "$1=$stateValues{$_}";
        }
    }
    join("; ", @cookie);
}

# Check a received reply for consistency, and remember necessary
# items in state file.
sub stateCheckReply {
    my $parsedReply = shift;
    if (!exists $parsedReply->{rst}) {
        die "ERROR: no result file received\n";
    }
    if (!$parsedReply->{rst}{player}{raceid}) {
        die "ERROR: result does not contain player name\n";
    }
    # Since 2012/02/10. the player objects no longer contain savekeys.
    # if ($parsedReply->{rst}{player}{savekey} ne $parsedReply->{savekey}) {
    #     die "ERROR: received two different savekeys\n"
    # }
    stateSet('savekey', $parsedReply->{savekey});
    stateSet('player', $parsedReply->{rst}{player}{raceid});     # user-visible player number ("7 = crystal")
    stateSet('playerid', $parsedReply->{rst}{player}{id});       # internal player Id used for everything
    stateSet('turn', $parsedReply->{rst}{settings}{turn});
}

sub doStatus {
    foreach (sort keys %stateValues) {
        my $v = stateQuote($stateValues{$_});
        print "$_ =\n";
        if (length($v) > 70) {
            print "   ", substr($v, 0, 67), "...\n";
        } else {
            print "   ", $v, "\n";
        }
    }
}

######################################################################
#
#  HTTP
#
######################################################################

sub httpCall {
    # Prepare
    my ($head, $body) = @_;
    my $host = stateGet('api');
    my $keks = stateCookies();
    $head .= "Host: $host\n";
    $head .= "Content-Length: " . length($body) . "\n";
    $head .= "Connection: close\n";
    $head .= "Cookie: $keks\n" if $keks ne '';
    $head .= "Content-Type: application/x-www-form-urlencoded; charset=UTF-8\n" if $body ne '';
    # $head .= "User-Agent: $0\n";
    $head =~ s/\n/\r\n/;
    $head .= "\r\n";

    # Socket cruft
    print "Calling server...\n";
    my $ip = inet_aton($host) or die "ERROR: unable to resolve host '$host': $!\n";
    my $paddr = sockaddr_in(80, $ip);
    socket(HTTP, PF_INET, SOCK_STREAM, getprotobyname('tcp')) or die "ERROR: unable to create socket: $!\n";
    binmode HTTP;
    HTTP->autoflush(1);
    connect(HTTP, $paddr) or die "ERROR: unable to connect to '$host': $!\n";

    # print "\033[36m$head$body\033[0m\n";

    # Send request
    print HTTP $head, $body;

    # Read reply header
    my %reply;
    while (<HTTP>) {
        s/[\r\n]*$//;
        if (/^$/) {
            last
        } elsif (m|^HTTP/\d+\.\d+ (\d+)|) {
            $reply{STATUS} = $1;
        } elsif (m|^set-cookie:\s*(.*?)=(.*?);|i) {
            $reply{"COOKIE-$1"} = $2;
        } elsif (m|^(.*?):\s+(.*)|) {
            $reply{lc($1)} = $2;
        } else {
            print STDERR "Unable to parse reply line '$_'\n";
        }
    }

    # Read reply body
    my $replybody;
    if (exists $reply{'content-length'}) {
        read HTTP, $replybody, $reply{'content-length'}
    } else {
        while (1) {
            my $tmp;
            if (!read(HTTP, $tmp, 4096)) { last }
            $replybody .= $tmp;
        }
    }
    close HTTP;

    # Check status
    if ($reply{STATUS} != 200) {
        print STDERR "WARNING: HTTP status is $reply{STATUS}.\n";
    }

    # Body might be compressed; decompress it
    if (exists $reply{'content-encoding'} && lc($reply{'content-encoding'}) eq 'gzip') {
        print "Decompressing result...\n";
        open TMP, "> c2nu.gz" or die "Cannot open temporary file: $!\n";
        binmode TMP;
        print TMP $replybody;
        close TMP;
        $replybody = "";
        open TMP, "gzip -dc c2nu.gz |" or die "Cannot open gzip: $!\n";
        binmode TMP;
        while (1) {
            my $tmp;
            if (!read(TMP, $tmp, 4096)) { last }
            $replybody .= $tmp;
        }
        close TMP;
    }

    $reply{BODY} = $replybody;
    \%reply;
}

sub httpEscape {
    my $x = shift;
    $x =~ s/([&+%\r\n])/sprintf("%%%02X", ord($1))/eg;
    $x =~ s/ /+/g;
    $x;
}

sub httpBuildQuery {
    my @list;
    while (@_) {
        my $key = shift @_;
        my $val = shift @_;
        push @list, "$key=" . httpEscape($val);
    }
    join('&', @list);
}

######################################################################
#
#  JSON
#
######################################################################

sub jsonParse {
    my $str = shift;
    pos($str) = 0;
    jsonParse1(\$str);
}

sub jsonParse1 {
    my $pstr = shift;
    $$pstr =~ m|\G\s*|sgc;
    if ($$pstr =~ m#\G"(([^\\"]+|\\.)*)"#gc) {
        my $s = $1;
        $s =~ s|\\(.)|stateUnquote($1)|eg;
        # Nu data is in UTF-8. Translate what we can to latin-1, because
        # PCC does not handle UTF-8 in game files. Doing it here conveniently
        # handles all places with possible UTF-8, including ship names,
        # messages, and notes.
        utf8ToLatin1($s);
    } elsif ($$pstr =~ m|\G([-+]?\d+\.\d*)|gc) {
        $1;
    } elsif ($$pstr =~ m|\G([-+]?\.\d+)|gc) {
        $1;
    } elsif ($$pstr =~ m|\G([-+]?\d+)|gc) {
        $1;
    } elsif ($$pstr =~ m|\Gtrue\b|gc) {
        1
    } elsif ($$pstr =~ m|\Gfalse\b|gc) {
        0
    } elsif ($$pstr =~ m|\Gnull\b|gc) {
        undef
    } elsif ($$pstr =~ m|\G\{|gc) {
        my $result = {};
        while (1) {
            $$pstr =~ m|\G\s*|sgc;
            if ($$pstr =~ m|\G\}|gc) { last }
            elsif ($$pstr =~ m|\G,|gc) { }
            else {
                my $key = jsonParse1($pstr);
                $$pstr =~ m|\G\s*|sgc;
                if ($$pstr !~ m|\G:|gc) { die "JSON syntax error: expecting ':', got '" . substr($$pstr, pos($$pstr), 20) . "'.\n" }
                my $val = jsonParse1($pstr);
                $result->{$key} = $val;
            }
        }
        $result;
    } elsif ($$pstr =~ m|\G\[|gc) {
        my $result = [];
        while (1) {
            $$pstr =~ m|\G\s*|sgc;
            if ($$pstr =~ m|\G\]|gc) { last }
            elsif ($$pstr =~ m|\G,|gc) { }
            else { push @$result, jsonParse1($pstr) }
        }
        $result;
    } else {
        die "JSON syntax error: expecting element, got '" . substr($$pstr, pos($$pstr), 20) . "'.\n";
    }
}

######################################################################
#
#  Utilities
#
######################################################################

sub replicate {
    my $n = shift;
    my @result;
    foreach (1 .. $n) { push @result, @_ }
    @result;
}

sub utf8ToLatin1 {
    my $s = shift;
    $s =~ s/([\xC0-\xC3])([\x80-\xBF])/chr(((ord($1) & 3) << 6) + (ord($2) & 63))/eg;
    $s;
}

sub readFile {
    my $f = shift;
    open IN, "< $f" or die "$f: $!\n";
    my $body;
    while (1) {
        my $tmp;
        if (!read(IN, $tmp, 4096)) { last }
        $body .= $tmp;
    }
    close IN;
    $body;
}
